# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/9/20 19:36
# Description: 
# ==============================================
import functools


def timeit(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        import time
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f'{func.__name__} run time: {end - start}s')
        return result

    return wrapper


def avg_time(func):
    whole_time: int = 0
    count: int = 0

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        import time
        nonlocal whole_time
        nonlocal count
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        whole_time += end - start
        count += 1
        print(f'{func.__name__} avg run time: {whole_time / count}s, call count: {count}')
        return result

    return wrapper



# -*- encoding:utf-8 -*-
# https://wiki.python.org/moin/PyQt/Python%20syntax%20highlighting

GLOBAL_STYLESHEET2 = """
QPushButton {
    border: 1px solid rgba(31, 35, 40, 0.15);
    border-radius: 2px;
    padding-left: 8px;
    padding-right: 8px;
    background-color: rgb(246, 248, 250);
    font: 12px "Microsoft Yahei";
    color: rgb(36, 41, 47);
}

QPushButton:hover {
    border: 1px solid rgb(221, 221, 221);
}
"""

GLOBAL_STYLESHEET = """
* {
    font-family: "Microsoft Yahei";
}

QMainWindow {
    background-color: #F0F0F0;
}

QPushButton {
    background-color: #FAFBFC;
    border: 1px solid rgba(27, 31, 35, 0.15);
    border-radius: 3px;
    color: #24292E;
    font: 12px "Microsoft Yahei";
    padding: 6px 10px;
}

QPushButton:hover {
    background-color: #F3F4F6;
}

QPushButton:pressed {
    background-color: #EDEFF2;
}

QPushButton:checked {
    background-color: #EDEFF2;
}

QPushButton:disabled {
    background-color: #FAFBFC;
    border-color: rgba(27, 31, 35, 0.15);
    color: #959DA5;
}

QLineEdit {
    border: 1px solid #D1D5DA;
    border-radius: 3px;
    padding: 6px 8px;
    font: 12px "Microsoft Yahei";
    color: #24292E;
}

QLineEdit:hover {
    border-color: #0366D6;
}

QLineEdit:focus {
    border-color: #0366D6;
    outline: none;
}

QLineEdit:disabled {
    background-color: #F0F0F0;
    border-color: rgba(27, 31, 35, 0.15);
    color: #959DA5;
}

QLineEdit#codeStyleLineEdit {
    font: 13px "Consolas";
}

QListWidget#stacksListWidget {
    border: 1px solid #D1D5DA;
    border-radius: 3px;
    font: 13px "Consolas";
    color: #24292E;
}

QListWidget#stacksListWidget::item {
    padding: 2px 0px;
}

QGroupBox {
    font: 13px "Microsoft Yahei";
}
"""



import os
import sys


def get_main_thread_instance(threading):
    if hasattr(threading, 'main_thread'):
        return threading.main_thread()
    else:
        # On Python 2 we don't really have an API to get the main thread,
        # so, we just get it from the 'shutdown' bound method.
        return threading._shutdown.im_self


def get_main_thread_id(unlikely_thread_id=None):
    '''
    :param unlikely_thread_id:
        Pass to mark some thread id as not likely the main thread.

    :return tuple(thread_id, critical_warning)
    '''
    import sys
    import os

    current_frames = sys._current_frames()
    possible_thread_ids = []
    for thread_ident, frame in current_frames.items():
        while frame.f_back is not None:
            frame = frame.f_back

        basename = os.path.basename(frame.f_code.co_filename)
        if basename.endswith(('.pyc', '.pyo')):
            basename = basename[:-1]

        if (frame.f_code.co_name, basename) in [
            ('_run_module_as_main', 'runpy.py'),
            ('run_module_as_main', 'runpy.py'),
            ('run_module', 'runpy.py'),
            ('run_path', 'runpy.py'),
        ]:
            # This is the case for python -m <module name> (this is an ideal match, so,
            # let's return it).
            return thread_ident, ''

        if frame.f_code.co_name == '<module>':
            if frame.f_globals.get('__name__') == '__main__':
                possible_thread_ids.insert(0, thread_ident)  # Add with higher priority
                continue

            # Usually the main thread will be started in the <module>, whereas others would
            # be started in another place (but when Python is embedded, this may not be
            # correct, so, just add to the available possibilities as we'll have to choose
            # one if there are multiple).
            possible_thread_ids.append(thread_ident)

    if len(possible_thread_ids) > 0:
        if len(possible_thread_ids) == 1:
            return possible_thread_ids[0], ''  # Ideal: only one match

        while unlikely_thread_id in possible_thread_ids:
            possible_thread_ids.remove(unlikely_thread_id)

        if len(possible_thread_ids) == 1:
            return possible_thread_ids[0], ''  # Ideal: only one match

        elif len(possible_thread_ids) > 1:
            # Bad: we can't really be certain of anything at this point.
            return possible_thread_ids[0], \
                'Multiple thread ids found (%s). Choosing main thread id randomly (%s).' % (
                    possible_thread_ids, possible_thread_ids[0])

    # If we got here we couldn't discover the main thread id.
    return None, 'Unable to discover main thread id.'


def fix_main_thread_id(on_warn=lambda msg: None, on_exception=lambda msg: None, on_critical=lambda msg: None):
    # This means that we weren't able to import threading in the main thread (which most
    # likely means that the main thread is paused or in some very long operation).
    # In this case we'll import threading here and hotfix what may be wrong in the threading
    # module (if we're on Windows where we create a thread to do the attach and on Linux
    # we are not certain on which thread we're executing this code).
    #
    # The code below is a workaround for https://bugs.python.org/issue37416
    import sys
    import threading

    try:
        with threading._active_limbo_lock:
            main_thread_instance = get_main_thread_instance(threading)

            if sys.platform == 'win32':
                # On windows this code would be called in a secondary thread, so,
                # the current thread is unlikely to be the main thread.
                if hasattr(threading, '_get_ident'):
                    unlikely_thread_id = threading._get_ident()  # py2
                else:
                    unlikely_thread_id = threading.get_ident()  # py3
            else:
                unlikely_thread_id = None

            main_thread_id, critical_warning = get_main_thread_id(unlikely_thread_id)

            if main_thread_id is not None:
                main_thread_id_attr = '_ident'
                if not hasattr(main_thread_instance, main_thread_id_attr):
                    main_thread_id_attr = '_Thread__ident'
                    assert hasattr(main_thread_instance, main_thread_id_attr)

                if main_thread_id != getattr(main_thread_instance, main_thread_id_attr):
                    # Note that we also have to reset the '_tstack_lock' for a regular lock.
                    # This is needed to avoid an error on shutdown because this lock is bound
                    # to the thread state and will be released when the secondary thread
                    # that initialized the lock is finished -- making an assert fail during
                    # process shutdown.
                    main_thread_instance._tstate_lock = threading._allocate_lock()
                    main_thread_instance._tstate_lock.acquire()

                    # Actually patch the thread ident as well as the threading._active dict
                    # (we should have the _active_limbo_lock to do that).
                    threading._active.pop(getattr(main_thread_instance, main_thread_id_attr), None)
                    setattr(main_thread_instance, main_thread_id_attr, main_thread_id)
                    threading._active[getattr(main_thread_instance, main_thread_id_attr)] = main_thread_instance

        # Note: only import from pydevd after the patching is done (we want to do the minimum
        # possible when doing that patching).
        on_warn(
            'The threading module was not imported by user code in the main thread. The debugger will attempt to work around https://bugs.python.org/issue37416.')

        if critical_warning:
            on_critical(
                'Issue found when debugger was trying to work around https://bugs.python.org/issue37416:\n%s' % (
                critical_warning,))
    except:
        on_exception('Error patching main thread id.')


def _try_get_qtlib_type():
    for module_name in sys.modules:
        if module_name.startswith('PyQt5'):
            return 'pyqt5'
        elif module_name.startswith('PySide2'):
            return 'pyside2'
        elif module_name.startswith('PyQt6'):
            return 'pyqt6'
        elif module_name.startswith('PySide6'):
            return 'pyside6'

    # No Qt libraries are imported
    # Try to import
    try:
        import PyQt5
        return 'pyqt5'
    except ImportError:
        try:
            import PySide2
            return 'pyside2'
        except ImportError:
            try:
                import PyQt6
                return 'pyqt6'
            except ImportError:
                try:
                    import PySide6
                    return 'pyside6'
                except ImportError:
                    pass

    # No Qt libraries are supported
    return ''


def attach(port, host, protocol=''):
    try:
        import sys
        # fix_main_thread = 'threading' not in sys.modules
        #
        # if fix_main_thread:
        #
        #     def on_warn(msg):
        #         from _pydev_bundle import pydev_log
        #         pydev_log.warn(msg)
        #
        #     def on_exception(msg):
        #         from _pydev_bundle import pydev_log
        #         pydev_log.error(msg)
        #
        #     def on_critical(msg):
        #         from _pydev_bundle import pydev_log
        #         pydev_log.error(msg)
        #
        #     fix_main_thread_id(on_warn=on_warn, on_exception=on_exception, on_critical=on_critical)
        #
        # else:
        #     from _pydev_bundle import pydev_log  # @Reimport
        #     pydev_log.debug('The threading module is already imported by user code.')
        #
        # if protocol:
        #     from _pydevd_bundle import pydevd_defaults
        #     pydevd_defaults.PydevdCustomization.DEFAULT_PROTOCOL = protocol

        import PyQtInspect.pqi

        # I.e.: disconnect/reset if already connected.

        PyQtInspect.pqi.SetupHolder.setup = None

        py_db = PyQtInspect.pqi.get_global_debugger()
        if py_db is not None:
            py_db.dispose_and_kill_all_pqi_threads()

        # pydevd.DebugInfoHolder.DEBUG_RECORD_SOCKET_READS = True
        # pydevd.DebugInfoHolder.LOG_TO_FILE_LEVEL = 3
        qtlib_type = _try_get_qtlib_type()
        PyQtInspect.pqi.settrace(
            port=port,
            host=host,
            patch_multiprocessing=True,
            is_attach=True,
            qt_support=qtlib_type,
        )
    except:
        import traceback
        traceback.print_exc()



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/9/7 16:11
# Description: 
# ==============================================
import sys

from PyQt5 import QtWidgets, QtGui, QtCore

settingsFile = "settings.ini"

setting = QtCore.QSettings(settingsFile, QtCore.QSettings.IniFormat)
setting.setIniCodec("UTF-8")

_PYCHARM_EXECUTABLE_NAMES = ["pycharm64.exe", "pycharm.exe", "pycharm"]


def getPyCharmPath():
    return setting.value("PyCharmPath", "")


def setPyCharmPath(path: str):
    setting.setValue("PyCharmPath", path)
    setting.sync()


def findDefaultPycharmPath():
    import os, subprocess

    def findForWindows():
        """ for Windows, we can use powershell command to find the path """
        output = subprocess.run(
            'powershell -Command "$(Get-Command pycharm).path"',
            shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            encoding='utf-8'
        )
        if output.stdout:
            return output.stdout.strip()

    def findForUnix():
        """ for Unix-like systems, we can use which command to find the path """
        output = subprocess.run(
            'which pycharm',
            shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            encoding='utf-8'
        )
        if output.stdout:
            return output.stdout.strip()

    # First, try to use terminal command to find the path
    if sys.platform == "win32":
        defaultPath = findForWindows()
        if defaultPath:
            return defaultPath
    else:
        defaultPath = findForUnix()
        if defaultPath:
            return defaultPath

    # If the above method fails, we can try to find the path from the environment variables
    for path in os.environ["PATH"].split(";" if sys.platform == "win32" else ":"):
        for pycharm_exe_name in _PYCHARM_EXECUTABLE_NAMES:
            pycharm_path = os.path.join(path, pycharm_exe_name)
            if os.path.isfile(pycharm_path):
                return pycharm_path
    return ""



# -*- encoding:utf-8 -*-

from PyQt5 import QtWidgets, QtCore


class SimpleSettingLineEdit(QtWidgets.QWidget):
    def __init__(self, parent, key: str, defaultValue: str = ""):
        super().__init__(parent)
        self.setFixedHeight(32)

        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(5, 0, 5, 0)
        self._layout.setSpacing(10)

        self._keyLabel = QtWidgets.QLabel(self)
        self._keyLabel.setText(key)
        self._keyLabel.setAlignment(QtCore.Qt.AlignCenter)
        self._keyLabel.setWordWrap(True)
        self._keyLabel.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        self._layout.addWidget(self._keyLabel)

        self._valueLineEdit = QtWidgets.QLineEdit(self)
        self._valueLineEdit.setText(defaultValue)
        self._valueLineEdit.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)

        self._layout.addWidget(self._valueLineEdit)

    def setValue(self, value: str):
        self._valueLineEdit.setText(value)

    def getValue(self) -> str:
        return self._valueLineEdit.text()



# PQI Tools for Qt

# ==== TODO ====
# 这个最好写个单元测试代码
def find_name_in_mro(cls, name, default):
    """ Emulate _PyType_Lookup() in Objects/typeobject.c """
    for base in cls.__mro__:
        if hasattr(base, name):
            yield base, getattr(base, name)
    yield default, default


def find_callable_var(obj, name):
    null = object()
    for cls, cls_var in find_name_in_mro(type(obj), name, null):
        if callable(cls_var):
            return cls_var
    raise AttributeError(name)


def find_method_by_name_and_call(obj, name, *args, **kwargs):
    if callable(getattr(obj, name)):
        return getattr(obj, name)(*args, **kwargs)
    else:
        # Sometimes, ``obj`` has a variable with the same name as the method
        return find_callable_var(obj, name)(obj, *args, **kwargs)


def get_widget_class_name(widget):
    return widget.__class__.__name__


def get_widget_object_name(widget):
    return find_method_by_name_and_call(widget, 'objectName')


def get_widget_size(widget):
    size = find_method_by_name_and_call(widget, 'size')
    return size.width(), size.height()


def get_widget_pos(widget):
    pos = find_method_by_name_and_call(widget, 'pos')
    return pos.x(), pos.y()


def get_widget_parent(widget):
    return find_method_by_name_and_call(widget, 'parent')


def get_parent_info(widget):
    while True:
        try:
            parent = get_widget_parent(widget)
        except:
            break

        if parent is None:
            break
        widget = parent
        yield get_widget_class_name(widget), id(widget), get_widget_object_name(widget)


def get_stylesheet(widget):
    return find_method_by_name_and_call(widget, 'styleSheet')


def get_children_info(widget):
    children = find_method_by_name_and_call(widget, 'children')
    for child in children:
        yield get_widget_class_name(child), id(child), get_widget_object_name(child)


def import_Qt(qt_type: str):
    """
    Import Qt libraries by type.

    :param qt_type: The Qt type to import, either 'pyqt5' or 'pyside2'.
    """
    if qt_type == 'pyqt5':
        import PyQt5 as QtLib
    elif qt_type == 'pyqt6':
        import PyQt6 as QtLib
    elif qt_type == 'pyside2':
        import PySide2 as QtLib
    elif qt_type == 'pyside6':
        import PySide6 as QtLib
    else:
        raise ValueError(f'Unsupported Qt type: {qt_type}')

    return QtLib


def import_wrap_module(qt_type: str):
    """
    Import the wrap module by Qt type.

    :param qt_type: The Qt type to import, either 'pyqt5' or 'pyside2'.
    """
    if qt_type == 'pyqt5':
        from PyQt5 import sip as wrap_module
        wrap_module._pqi_is_valid = lambda x: wrap_module.isdeleted(x) == False
    elif qt_type == 'pyqt6':
        from PyQt6 import sip as wrap_module
        wrap_module._pqi_is_valid = lambda x: wrap_module.isdeleted(x) == False
    elif qt_type == 'pyside2':
        import shiboken2 as wrap_module
        wrap_module._pqi_is_valid = wrap_module.isValid
    elif qt_type == 'pyside6':
        import shiboken6 as wrap_module
        wrap_module._pqi_is_valid = wrap_module.isValid
    else:
        raise ValueError(f'Unsupported Qt type: {qt_type}')

    return wrap_module


def _send_custom_event(target_widget, key: str, val):
    from PyQtInspect.pqi import SetupHolder

    QtCore = import_Qt(SetupHolder.setup['qt-support']).QtCore
    EventEnum = QtCore.QEvent.Type if hasattr(QtCore.QEvent, 'Type') else QtCore.QEvent
    event = QtCore.QEvent(EventEnum.User)
    setattr(event, key, val)
    QtCore.QCoreApplication.postEvent(target_widget, event)


def set_widget_highlight(widget, highlight: bool):
    """
    Set the highlight on a widget.

    :note: Use custom events to avoid program crashes due to cross-threaded calls

    :param widget: The widget to set the highlight on.

    :param highlight: A boolean indicating whether to highlight the widget or not.
    """
    _send_custom_event(widget, '_pqi_is_highlight', highlight)


def exec_code_in_widget(widget, code: str):
    _send_custom_event(widget, '_pqi_exec_code', code)


def is_wrapped_pointer_valid(ptr):
    """
    Check if a wrapped pointer is valid.

    :param ptr: The pointer to check.
    """
    from PyQtInspect.pqi import SetupHolder
    return import_wrap_module(SetupHolder.setup['qt-support'])._pqi_is_valid(ptr)



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2024/8/19 21:24
# Description:
# ==============================================
import abc

from PyQt5 import QtCore


class KeyboardHookHandler(QtCore.QObject):
    sigDisableInspectKeyPressed = QtCore.pyqtSignal()

    @classmethod
    def __new__(cls, *args, **kwargs):
        """ Factory method to create a proper instance according to the platform. """

        import sys
        if sys.platform == 'win32':
            class_ = WindowsKeyboardHookHandler
        else:
            class_ = DummyKeyboardHookHandler
        ins = super().__new__(class_)
        return ins

    def __init__(self, parent=None):
        super().__init__(parent)

    def isValid(self) -> bool:
        return True

    @abc.abstractmethod
    def onInspectFinished(self): ...

    @abc.abstractmethod
    def onInspectBegin(self): ...

    @abc.abstractmethod
    def onInspectDisabled(self): ...

    @abc.abstractmethod
    def setEnable(self, enable: bool): ...


class WindowsKeyboardHookHandler(KeyboardHookHandler):
    """ Keyboard hook handler for Windows platform. """
    sigKeyboardEvent = QtCore.pyqtSignal(int, int, int, int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._keyboardHookThread = self._generateKeyboardHookThread()
        self._enabled = True  # default enabled
        self._inspecting = False

    def _generateKeyboardHookThread(self):
        """ Initialize a thread to grab keyboard event. """
        from PyQtInspect._pqi_bundle.pqi_keyboard_hook_win import GrabFlag

        def _inSubThread():
            import PyQtInspect._pqi_bundle.pqi_keyboard_hook_win as kb_hook
            kb_hook.grab(0x77, flag, lambda: self.sigDisableInspectKeyPressed.emit())

        thread = QtCore.QThread(self)
        flag = GrabFlag()  # 可指示内部grab线程停止的标志, 用于解决GUI程序退出后grab线程仍无法退出的问题
        thread.started.connect(_inSubThread)
        thread.stop_flag = flag
        return thread

    def _stopKeyboardHookThread(self):
        if self._keyboardHookThread.isRunning():
            self._keyboardHookThread.stop_flag.mark_stop()
            self._keyboardHookThread.quit()

    def _onKeyboardEvent(self, nCode, wParam, lParam):
        self.sigKeyboardEvent.emit(nCode, wParam, lParam)

    def _startKeyboardHookThread(self):
        # start keyboard hook thread if user wants to disable inspect by pressing F8
        # TODO: 1) 自定义热键; 2) 当用户打开开关时, 且处于inspect, 立即运行线程
        self._keyboardHookThread.stop_flag.clear_flag()
        self._keyboardHookThread.start()

    # region Override methods
    def onInspectFinished(self):
        self._inspecting = False
        self._stopKeyboardHookThread()

    def onInspectBegin(self):
        self._inspecting = True
        if not self._enabled:
            return
        self._startKeyboardHookThread()

    def onInspectDisabled(self):
        self._inspecting = False
        self._stopKeyboardHookThread()

    def setEnable(self, enable: bool):
        self._enabled = enable
        if not enable:
            self._stopKeyboardHookThread()
        elif self._inspecting:  # enable and inspecting
            # start keyboard hook thread during inspecting
            self._startKeyboardHookThread()
    # endregion


class DummyKeyboardHookHandler(KeyboardHookHandler):
    """  For non-windows platform, do nothing. """
    def _dummy(self): ...

    onInspectFinished = onInspectBegin = onInspectDisabled = setEnable = _dummy

    def isValid(self) -> bool:
        return False


# -*- encoding:utf-8 -*-


class _DummyLogger:
    def _func(self, *args, **kwargs): pass

    def __getattr__(self, item):
        return self._func


def _get_trace_level():
    from PyQtInspect._pqi_bundle.pqi_contants import DebugInfoHolder

    return DebugInfoHolder.LOG_TO_FILE_LEVEL, DebugInfoHolder.LOG_TO_CONSOLE_LEVEL


def get_logger():
    from PyQtInspect._pqi_common.pqi_setup_holder import SetupHolder

    if SetupHolder.setup is None:
        return _DummyLogger

    if SetupHolder.setup.get('server'):
        from ._server import logger
    else:
        from ._client import logger

    file_log_level, console_log_level = _get_trace_level()
    logger.set_console_log_level(console_log_level)
    logger.set_file_log_level(file_log_level)
    return logger


def debug(msg, *args, **kwargs):
    get_logger().debug(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    get_logger().info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    get_logger().warning(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    get_logger().error(msg, *args, **kwargs)


def critical(msg, *args, **kwargs):
    get_logger().critical(msg, *args, **kwargs)



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/8/18 14:24
# Description: 
# ==============================================
'''
This module holds the constants used for specifying the states of the debugger.
'''
from __future__ import nested_scopes

import logging
import platform
import sys  # Note: the sys import must be here anyways (others depend on it)

STATE_RUN = 1
STATE_SUSPEND = 2

PYTHON_SUSPEND = 1
DJANGO_SUSPEND = 2
JINJA2_SUSPEND = 3
JUPYTER_SUSPEND = 4

GET_FRAME_NORMAL_GROUP = 0
GET_FRAME_SPECIAL_GROUP = 1
GET_FRAME_RETURN_GROUP = 2


class DebugInfoHolder:
    # we have to put it here because it can be set through the command line (so, the
    # already imported references would not have it).
    DEBUG_RECORD_SOCKET_READS = False
    LOG_TO_FILE_LEVEL = logging.INFO
    LOG_TO_CONSOLE_LEVEL = logging.WARNING


IS_CPYTHON = platform.python_implementation() == 'CPython'

# Used to determine the maximum size of each variable passed to eclipse -- having a big value here may make
# the communication slower -- as the variables are being gathered lazily in the latest version of eclipse,
# this value was raised from 200 to 1000.
MAXIMUM_VARIABLE_REPRESENTATION_SIZE = 1000
# Prefix for saving functions return values in locals
RETURN_VALUES_DICT = '__pydevd_ret_val_dict'

original_excepthook = sys.__excepthook__


def dummy_excepthook(exctype, value, traceback):
    return None


import os

# Constant detects when running on Jython/windows properly later on.
IS_WINDOWS = sys.platform == 'win32'

IS_64BIT_PROCESS = sys.maxsize > (2 ** 32)

IS_LINUX = sys.platform.startswith('linux')
IS_MACOS = sys.platform == 'darwin'

IS_PYTHON_STACKLESS = "stackless" in sys.version.lower()
CYTHON_SUPPORTED = False

NUMPY_NUMERIC_TYPES = "biufc"
NUMPY_FLOATING_POINT_TYPES = "fc"
# b boolean
# i signed integer
# u unsigned integer
# f floating-point
# c complex floating-point

try:
    import platform
    python_implementation = platform.python_implementation()
except:
    pass
else:
    if python_implementation == 'CPython' and not IS_PYTHON_STACKLESS:
        # Only available for CPython!
        if (
            (sys.version_info[0] == 2 and sys.version_info[1] >= 7)
            or (sys.version_info[0] == 3 and sys.version_info[1] >= 5)
            or (sys.version_info[0] > 3)
        ):
            # Supported in 2.7 or 3.5 onwards (32 or 64)
            CYTHON_SUPPORTED = True

IS_PYCHARM_ATTACH = os.getenv('PYCHARM_ATTACH') == 'True'

#=======================================================================================================================
# Python 3?
#=======================================================================================================================
IS_PY3K = False
IS_PY34_OR_GREATER = False
IS_PY36_OR_GREATER = False
IS_PY37_OR_GREATER = False
IS_PY36_OR_LESSER = False
IS_PY38_OR_GREATER = False
IS_PY38 = False
IS_PY39_OR_GREATER = False
IS_PY311 = False
IS_PY311_OR_GREATER = False
IS_PY2 = True
IS_PY27 = False
IS_PY24 = False
try:
    if sys.version_info[0] >= 3:
        IS_PY3K = True
        IS_PY2 = False
        IS_PY34_OR_GREATER = sys.version_info >= (3, 4)
        IS_PY36_OR_GREATER = sys.version_info >= (3, 6)
        IS_PY37_OR_GREATER = sys.version_info >= (3, 7)
        IS_PY36_OR_LESSER = sys.version_info[:2] <= (3, 6)
        IS_PY38 = sys.version_info[0] == 3 and sys.version_info[1] == 8
        IS_PY38_OR_GREATER = sys.version_info >= (3, 8)
        IS_PY39_OR_GREATER = sys.version_info >= (3, 9)
        IS_PY311 = sys.version_info[0] == 3 and sys.version_info[1] == 11
        IS_PY311_OR_GREATER = sys.version_info >= (3, 11)
    elif sys.version_info[0] == 2 and sys.version_info[1] == 7:
        IS_PY27 = True
    elif sys.version_info[0] == 2 and sys.version_info[1] == 4:
        IS_PY24 = True
except AttributeError:
    pass  # Not all versions have sys.version_info

try:
    SUPPORT_GEVENT = os.getenv('GEVENT_SUPPORT', 'False') == 'True'
except:
    # Jython 2.1 doesn't accept that construct
    SUPPORT_GEVENT = False

try:
    DROP_INTO_DEBUGGER_ON_FAILED_TESTS = os.environ.get('DROP_INTO_DEBUGGER_ON_FAILED_TESTS', 'False') == 'True'
except:
    DROP_INTO_DEBUGGER_ON_FAILED_TESTS = False

# At the moment gevent supports Python >= 2.6 and Python >= 3.3
USE_LIB_COPY = SUPPORT_GEVENT and \
               ((not IS_PY3K and sys.version_info[1] >= 6) or
                (IS_PY3K and sys.version_info[1] >= 3))


class ValuesPolicy:
    SYNC = 0
    ASYNC = 1
    ON_DEMAND = 2


LOAD_VALUES_POLICY = ValuesPolicy.SYNC
if os.getenv('PYDEVD_LOAD_VALUES_ASYNC', 'False') == 'True':
    LOAD_VALUES_POLICY = ValuesPolicy.ASYNC
if os.getenv('PYDEVD_LOAD_VALUES_ON_DEMAND', 'False') == 'True':
    LOAD_VALUES_POLICY = ValuesPolicy.ON_DEMAND
DEFAULT_VALUES_DICT = {ValuesPolicy.ASYNC: "__pydevd_value_async", ValuesPolicy.ON_DEMAND: "__pydevd_value_on_demand"}

INTERACTIVE_MODE_AVAILABLE = sys.platform in ('darwin', 'win32') or os.getenv('DISPLAY') is not None
IS_PYCHARM = True
ASYNC_EVAL_TIMEOUT_SEC = 60
NEXT_VALUE_SEPARATOR = "__pydev_val__"
BUILTINS_MODULE_NAME = '__builtin__' if IS_PY2 else 'builtins'
SHOW_DEBUG_INFO_ENV = os.getenv('PYCHARM_DEBUG') == 'True' or os.getenv('PQI_DEBUG') == 'True'
IS_ASYNCIO_REPL = os.getenv('ASYNCIO_REPL') == 'True' and IS_PY38_OR_GREATER
IS_ASYNCIO_DEBUGGER_ENV = os.getenv('ASYNCIO_DEBUGGER_ENV') == 'True' and IS_PY38_OR_GREATER

# If True, CMD_SET_NEXT_STATEMENT and CMD_RUN_TO_LINE commands have responses indicating success or failure.
GOTO_HAS_RESPONSE = IS_PYCHARM

if SHOW_DEBUG_INFO_ENV:
    # show debug info before the debugger start
    DebugInfoHolder.DEBUG_RECORD_SOCKET_READS = True
    DebugInfoHolder.LOG_TO_FILE_LEVEL = logging.DEBUG
    DebugInfoHolder.LOG_TO_CONSOLE_LEVEL = logging.DEBUG


def protect_libraries_from_patching():
    """
    In this function we delete some modules from `sys.modules` dictionary and import them again inside
      `_pydev_saved_modules` in order to save their original copies there. After that we can use these
      saved modules within the debugger to protect them from patching by external libraries (e.g. gevent).
    """
    patched = ['threading', 'thread', '_thread', 'time', 'socket', 'Queue', 'queue', 'select',
               'xmlrpclib', 'SimpleXMLRPCServer', 'BaseHTTPServer', 'SocketServer',
               'xmlrpc.client', 'xmlrpc.server', 'http.server', 'socketserver']

    for name in patched:
        try:
            __import__(name)
        except:
            pass

    patched_modules = dict([(k, v) for k, v in sys.modules.items()
                            if k in patched])

    for name in patched_modules:
        del sys.modules[name]

    # import for side effects
    import PyQtInspect._pqi_imps._pqi_saved_modules

    for name in patched_modules:
        sys.modules[name] = patched_modules[name]


if USE_LIB_COPY:
    protect_libraries_from_patching()

from PyQtInspect._pqi_imps._pqi_saved_modules import thread
_thread_id_lock = thread.allocate_lock()
thread_get_ident = thread.get_ident

if IS_PY3K:

    def dict_keys(d):
        return list(d.keys())

    def dict_values(d):
        return list(d.values())

    dict_iter_values = dict.values

    def dict_iter_items(d):
        return d.items()

    def dict_items(d):
        return list(d.items())

else:
    def dict_keys(d):
        return d.keys()

    try:
        dict_iter_values = dict.itervalues
    except:
        try:
            dict_iter_values = dict.values  # Older versions don't have the itervalues
        except:

            def dict_iter_values(d):
                return d.values()

    try:
        dict_values = dict.values
    except:

        def dict_values(d):
            return d.values()

    def dict_iter_items(d):
        try:
            return d.iteritems()
        except:
            return d.items()

    def dict_items(d):
        return d.items()

try:
    xrange = xrange
except:
    # Python 3k does not have it
    xrange = range

try:
    import itertools
    izip = itertools.izip
except:
    izip = zip


try:
    enumerate
except:
    def enumerate(lst):
        ret = []
        i = 0
        for element in lst:
            ret.append((i, element))
            i += 1
        return ret

try:
    from StringIO import StringIO
except:
    from io import StringIO

NO_FTRACE = None

if sys.version_info[:2] in ((3, 3), (3, 4), (3, 5)) or sys.version_info < (2, 7, 12):

    def NO_FTRACE(frame, event, arg):
        # In Python < 2.7.12 and <= 3.5, if we're tracing a method, frame.f_trace may not be set
        # to None, it must always be set to a tracing function.
        # See: pydev_tests_python.test_tracing_gotchas.test_tracing_gotchas
        return None


#=======================================================================================================================
# get_pid
#=======================================================================================================================
def get_pid():
    try:
        return os.getpid()
    except AttributeError:
        try:
            # Jython does not have it!
            import java.lang.management.ManagementFactory  # @UnresolvedImport -- just for jython
            pid = java.lang.management.ManagementFactory.getRuntimeMXBean().getName()
            return pid.replace('@', '_')
        except:
            # ok, no pid available (will be unable to debug multiple processes)
            return '000001'


def clear_cached_thread_id(thread):
    with _thread_id_lock:
        try:
            if thread.__pydevd_id__ != 'console_main':
                # The console_main is a special thread id used in the console and its id should never be reset
                # (otherwise we may no longer be able to get its variables -- see: https://www.brainwy.com/tracker/PyDev/776).
                del thread.__pydevd_id__
        except AttributeError:
            pass


# Don't let threads be collected (so that id(thread) is guaranteed to be unique).
_thread_id_to_thread_found = {}


def _get_or_compute_thread_id_with_lock(thread, is_current_thread):
    with _thread_id_lock:
        # We do a new check with the lock in place just to be sure that nothing changed
        tid = getattr(thread, '__pydevd_id__', None)
        if tid is not None:
            return tid

        _thread_id_to_thread_found[id(thread)] = thread

        # Note: don't use thread.ident because a new thread may have the
        # same id from an old thread.
        pid = get_pid()
        tid = 'pid_%s_id_%s' % (pid, id(thread))

        thread.__pydevd_id__ = tid

    return tid


def get_current_thread_id(thread):
    '''
    Note: the difference from get_current_thread_id to get_thread_id is that
    for the current thread we can get the thread id while the thread.ident
    is still not set in the Thread instance.
    '''
    try:
        # Fast path without getting lock.
        tid = thread.__pydevd_id__
        if tid is None:
            # Fix for https://www.brainwy.com/tracker/PyDev/645
            # if __pydevd_id__ is None, recalculate it... also, use an heuristic
            # that gives us always the same id for the thread (using thread.ident or id(thread)).
            raise AttributeError()
    except AttributeError:
        tid = _get_or_compute_thread_id_with_lock(thread, is_current_thread=True)

    return tid


#=======================================================================================================================
# Null
#=======================================================================================================================
class Null:
    """
    Gotten from: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/68205
    """

    def __init__(self, *args, **kwargs):
        return None

    def __call__(self, *args, **kwargs):
        return self

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        return self

    def __getattr__(self, mname):
        if len(mname) > 4 and mname[:2] == '__' and mname[-2:] == '__':
            # Don't pretend to implement special method names.
            raise AttributeError(mname)
        return self

    def __setattr__(self, name, value):
        return self

    def __delattr__(self, name):
        return self

    def __repr__(self):
        return "<Null>"

    def __str__(self):
        return "Null"

    def __len__(self):
        return 0

    def __getitem__(self):
        return self

    def __setitem__(self, *args, **kwargs):
        pass

    def write(self, *args, **kwargs):
        pass

    def __nonzero__(self):
        return 0

    def __iter__(self):
        return iter(())


# Default instance
NULL = Null()


def call_only_once(func):
    '''
    To be used as a decorator

    @call_only_once
    def func():
        print 'Calling func only this time'

    Actually, in PyDev it must be called as:

    func = call_only_once(func) to support older versions of Python.
    '''

    def new_func(*args, **kwargs):
        if not new_func._called:
            new_func._called = True
            return func(*args, **kwargs)

    new_func._called = False
    return new_func


#=======================================================================================================================
# GlobalDebuggerHolder
#=======================================================================================================================
class GlobalDebuggerHolder:
    '''
        Holder for the global debugger.
    '''
    global_dbg = None  # Note: don't rename (the name is used in our attach to process)


#=======================================================================================================================
# get_global_debugger
#=======================================================================================================================
def get_global_debugger():
    return GlobalDebuggerHolder.global_dbg


GetGlobalDebugger = get_global_debugger  # Backward-compatibility


#=======================================================================================================================
# set_global_debugger
#=======================================================================================================================
def set_global_debugger(dbg):
    GlobalDebuggerHolder.global_dbg = dbg


# =======================================================================================================================
# Qt Patch
# =======================================================================================================================

QtWidgetClasses = [
    'QWidget',
    'QAbstractButton',
    'QFrame',
    'QAbstractSlider',
    'QAbstractSpinBox',
    'QCalendarWidget',
    'QDialog',
    'QComboBox',
    'QDesktopWidget',
    'QDialogButtonBox',
    'QDockWidget',
    'QFocusFrame',
    'QGroupBox',
    'QKeySequenceEdit',
    'QLineEdit',
    'QMainWindow',
    'QMdiSubWindow',
    'QMenu',
    'QMenuBar',
    'QOpenGLWidget',
    'QProgressBar',
    'QRubberBand',
    'QSizeGrip',
    'QSplashScreen',
    'QSplitterHandle',
    'QStatusBar',
    'QTabBar',
    'QTabWidget',
    'QToolBar',
    'QWizardPage',
    'QCheckBox',
    'QPushButton',
    'QRadioButton',
    'QToolButton',
    'QCommandLinkButton',
    'QAbstractScrollArea',
    'QLCDNumber',
    'QLabel',
    # 'QSplitter',  # ignored
    'QStackedWidget',
    'QToolBox',
    'QAbstractItemView',
    'QGraphicsView',
    'QMdiArea',
    'QPlainTextEdit',
    'QScrollArea',
    'QTextEdit',
    'QColumnView',
    'QHeaderView',
    'QListView',
    'QTableView',
    'QTreeView',
    'QListWidget',
    'QUndoView',
    'QTableWidget',
    'QTreeWidget',
    'QTextBrowser',
    'QDial',
    'QScrollBar',
    'QSlider',
    'QDateTimeEdit',
    'QDoubleSpinBox',
    'QSpinBox',
    'QDateEdit',
    'QTimeEdit',
    'QColorDialog',
    'QErrorMessage',
    'QFileDialog',
    'QFontDialog',
    'QInputDialog',
    'QMessageBox',
    'QProgressDialog',
    'QWizard',
    'QFontComboBox',
]

SUPPORT_CC = True

if __name__ == '__main__':
    if Null():
        sys.stdout.write('here\n')



# -*- encoding:utf-8 -*-
import collections
import sys
from contextlib import redirect_stdout
from ctypes import wintypes
from io import StringIO
import os

from PyQtInspect._pqi_bundle import pqi_log
from PyQtInspect._pqi_bundle.pqi_contants import get_global_debugger, QtWidgetClasses, IS_WINDOWS, IS_MACOS
from PyQtInspect._pqi_bundle.pqi_path_helper import find_pqi_module_path, is_relative_to
from PyQtInspect._pqi_bundle.pqi_qt_tools import get_widget_size
from PyQtInspect._pqi_bundle.pqi_stack_tools import getStackFrame
from PyQtInspect._pqi_bundle.pqi_log.log_utils import log_exception


_PQI_MOCKED_EVENT_ATTR = '_pqi_mocked'
_PQI_INSPECTED_PROP_NAME = '_pqi_inspected'
_PQI_INSPECTED_PROP_NAME_BYTES = b'_pqi_inspected'


def _get_filename_from_frame(frame):
    return frame.f_code.co_filename


def _try_get_file_name(frame):
    """
    :return: filename, is source code
    """
    try:
        # src_filename = inspect.getsourcefile(frame)
        return _get_filename_from_frame(frame)
    except TypeError:
        return '', False


def filter_trace_stack(traceStacks):
    filteredStacks = []
    from PyQtInspect.pqi import SetupHolder
    stackMaxDepth = SetupHolder.setup["stack-max-depth"]
    showPqiStack = SetupHolder.setup["show-pqi-stack"]
    pqi_module_path = find_pqi_module_path()
    stacks = traceStacks[2:stackMaxDepth + 1] if stackMaxDepth != 0 else traceStacks[2:]
    for frame, lineno in stacks:
        filename = _get_filename_from_frame(frame)
        if not showPqiStack and is_relative_to(filename, pqi_module_path):
            break
        filteredStacks.append(
            {
                'filename': filename,
                'lineno': lineno,
                'function': frame.f_code.co_name,
            }
        )
    return filteredStacks


def _is_inspect_enabled():
    debugger = get_global_debugger()
    return debugger is not None and debugger.inspect_enabled


def patch_QtWidgets(QtModule, qt_support_mode='auto', is_attach=False):
    QtWidgets = QtModule.QtWidgets
    QtGui = QtModule.QtGui
    QtCore = QtModule.QtCore

    EventEnum = QtCore.QEvent.Type if hasattr(QtCore.QEvent, 'Type') else QtCore.QEvent
    WidgetAttributeEnum = QtCore.Qt.WidgetAttribute if hasattr(QtCore.Qt, 'WidgetAttribute') else QtCore.Qt
    MouseButtonEnum = QtCore.Qt.MouseButton if hasattr(QtCore.Qt, 'MouseButton') else QtCore.Qt
    KeyboardModifierEnum = QtCore.Qt.KeyboardModifier if hasattr(QtCore.Qt, 'KeyboardModifier') else QtCore.Qt

    if qt_support_mode.startswith("pyqt"):
        sip = QtModule.sip
        isdeleted = sip.isdeleted
        ispycreated = sip.ispycreated
    elif qt_support_mode.startswith("pyside"):  # todo pyside6 also use this?
        if qt_support_mode == 'pyside2':
            import shiboken2 as _shiboken
        elif qt_support_mode == 'pyside6':
            import shiboken6 as _shiboken
        isdeleted = lambda obj: not _shiboken.isValid(obj)
        ispycreated = _shiboken.createdByPython

    def _create_mouse_event(event_type, pos, button):
        """ Create a mouse event with the specified parameters.
        It is safe, because the event object created by Python is allocated on the heap.
        """
        return QtGui.QMouseEvent(event_type, QtCore.QPointF(pos), button, button, KeyboardModifierEnum.NoModifier)

    def _register_widget(widget):
        debugger = get_global_debugger()
        if debugger is not None:
            debugger.register_widget(widget)

    def _createHighlightFg(parent: QtWidgets.QWidget):
        # 先new再调用之前的__init__
        widget = QtWidgets.QWidget.__new__(QtWidgets.QWidget)
        QtWidgets.QWidget._original_QWidget_init(widget, parent)
        widget.setFixedSize(*get_widget_size(parent))
        # 不要响应鼠标事件
        widget.setAttribute(WidgetAttributeEnum.WA_TransparentForMouseEvents)
        widget.setObjectName("_pqi_highlight_fg")
        widget.setStyleSheet("background-color: rgba(255, 0, 0, 0.2);")
        return widget

    def _mark_obj_inspected(obj):
        setattr(obj, '_pqi_inspected_mark', True)

    def _clear_obj_inspected_mark(obj):
        if hasattr(obj, '_pqi_inspected_mark'):
            del obj._pqi_inspected_mark

    def _is_obj_inspected(obj):
        return hasattr(obj, '_pqi_inspected_mark')

    class HighlightController:
        last_highlighted_widget = None
        # for some widgets like QSplitter, we should not highlight them, or they will change their size
        widget_class_to_ignore = (
            QtWidgets.QSplitter,
        )

        @classmethod
        def _is_ignored(cls, widget):
            return any(isinstance(widget, class_) for class_ in cls.widget_class_to_ignore)

        @classmethod
        def unhighlight_last(cls):
            if cls.last_highlighted_widget is not None and not isdeleted(cls.last_highlighted_widget):
                cls.last_highlighted_widget.hide()
            cls.last_highlighted_widget = None

        @classmethod
        def highlight(cls, widget):
            if cls._is_ignored(widget):
                return

            if not hasattr(widget, '_pqi_highlight_fg'):
                widget._pqi_highlight_fg = _createHighlightFg(widget)

            widget._pqi_highlight_fg.setFixedSize(*get_widget_size(widget))
            cls.unhighlight_last()
            widget._pqi_highlight_fg.show()
            cls.last_highlighted_widget = widget._pqi_highlight_fg

        @classmethod
        def unhighlight(cls, widget):
            if hasattr(widget, '_pqi_highlight_fg'):
                widget._pqi_highlight_fg.hide()
                if cls.last_highlighted_widget is widget._pqi_highlight_fg:
                    cls.last_highlighted_widget = None

    class EnteredWidgetStack:
        def __init__(self):
            self._stack = []

        def push(self, widget):
            self._stack.append(widget)

        def pop(self):
            self._stack.pop()

        def filter(self):
            while self._stack:
                wgt = self._stack[-1]
                if isdeleted(wgt):
                    self._stack.pop()
                else:
                    break

        def clear(self):
            self._stack.clear()

        def __bool__(self):
            return bool(self._stack)

        def __getitem__(self, item):
            return self._stack[item]

        def __len__(self):
            return len(self._stack)

    _entered_widget_stack = EnteredWidgetStack()

    def _inspect_widget(debugger, widget: QtWidgets.QWidget):
        # print('inspect:', widget.__class__.__name__, widget.objectName(), widget)
        # === send widget info === #
        debugger.send_widget_info_to_server(widget)

        # === highlight widget === #
        HighlightController.highlight(widget)

        # === hook mouseReleaseEvent === #
        _mark_obj_inspected(widget)

    def _inspect_top(stack: EnteredWidgetStack):
        stack.filter()
        if not stack:
            return

        if not _is_inspect_enabled():
            return
        debugger = get_global_debugger()

        obj = stack[-1]

        _inspect_widget(debugger, obj)

    class EventListener(QtCore.QObject):

        def _handleEnterEvent(self, obj, event):
            if not _is_inspect_enabled():
                return

            if _entered_widget_stack:
                # 如果栈顶有元素, 则取消掉栈顶元素的选择状态
                # 否则QTabWidget会表现异常
                # todo 看看能不能集成在栈上, 毕竟两者强关联
                _clear_obj_inspected_mark(_entered_widget_stack[-1])
            _entered_widget_stack.push(obj)
            _inspect_top(_entered_widget_stack)

        def _handleLeaveEvent(self, obj, event):
            # 注意这个不对称
            # 因为leaveEvent是在鼠标离开时触发的, 但是鼠标离开时, 有可能鼠标已经进入了下一个widget
            # 所以不能直接pop
            if not _is_inspect_enabled():
                return

            if _entered_widget_stack and _entered_widget_stack[-1] == obj:
                _entered_widget_stack.pop()
            else:
                _entered_widget_stack.clear()

            HighlightController.unhighlight(obj)
            _clear_obj_inspected_mark(obj)

            _inspect_top(_entered_widget_stack)

        def _handleMouseReleaseEvent(self, obj, event) -> bool:
            """ 处理鼠标点击事件, 这里需要返回一个bool值, 表示是否拦截事件 """
            if not _is_inspect_enabled():
                return False

            # print(f'click: {obj}, button: {event.button()}')
            if not _is_obj_inspected(obj):
                return False

            # 自己通过postEvent发出的事件, 不处理
            # 不使用event.spontaneous()判断, 因为对于QTextBrowser的点击事件, spontaneous()返回False
            if getattr(event, _PQI_MOCKED_EVENT_ATTR, False):
                return False

            debugger = get_global_debugger()

            if event.button() != MouseButtonEnum.LeftButton:
                if debugger is not None and debugger.mock_left_button_down and event.button() == MouseButtonEnum.RightButton:
                    # mock left button press and release event
                    # First, send a mouse press event
                    pressEvent = _create_mouse_event(EventEnum.MouseButtonPress, event.pos(), MouseButtonEnum.LeftButton)
                    # 使用postEvent传播事件, 而不是直接调用obj.mousePressEvent, 以便其他的eventFilter能够接收到这个事件
                    QtCore.QCoreApplication.postEvent(obj, pressEvent)

                    # Then, change the original event and send it again
                    event = _create_mouse_event(EventEnum.MouseButtonRelease, event.pos(), MouseButtonEnum.LeftButton)
                    setattr(event, _PQI_MOCKED_EVENT_ATTR, True)
                    # 同理, 为了能让后面的eventFilter能够接收到鼠标事件, 这里使用postEvent再次将事件传播出去
                    QtCore.QCoreApplication.postEvent(obj, event)
                    # stop event propagation
                    return True
                else:
                    # Bug Fixed 20240810: We CAN NOT re-post the original event,
                    # because it will be deleted after the event loop.
                    # see: https://doc.qt.io/qt-5/qcoreapplication.html#postEvent
                    # ---
                    # The event must be allocated on the heap
                    # since the post event queue will take ownership of the event
                    # and delete it once it has been posted.
                    # It is not safe to access the event after it has been posted.
                    # ---
                    # Note: all objects created in Python are allocated on the heap.
                    # see: https://docs.python.org/3/c-api/memory.html
                    # ---
                    return False

            # inspect finished
            debugger.notify_inspect_finished(obj)
            debugger.disable_inspect()
            HighlightController.unhighlight(obj)
            _entered_widget_stack.clear()
            _clear_obj_inspected_mark(obj)

            # stop event propagation
            return True

        def _handleMousePressEvent(self, obj, event):
            if not _is_inspect_enabled():
                return False
            # print(f'press: {obj}')
            if not event.spontaneous():
                return False
            if not _is_obj_inspected(obj):
                return False
            # obj.mousePressEvent(event)
            # 对于当前被检查的控件, 阻止MousePress事件传播
            # 防止点击后某些eventFilter在处理MousePress事件时使得inspect的控件改变, 导致后续的MouseRelease事件处理不正常
            return True

        def eventFilter(self, obj, event):
            # Intercept `QDynamicPropertyChange` events for properties dynamically
            # added by PyQtInspect itself (like `_pqi_inspected`).
            with log_exception(suppress=True):
                if (event.type() == EventEnum.DynamicPropertyChange
                        and bytes(event.propertyName()) == _PQI_INSPECTED_PROP_NAME_BYTES):
                    return True

                if not obj.property(_PQI_INSPECTED_PROP_NAME):
                    return False

                if event.type() == EventEnum.Enter:
                    self._handleEnterEvent(obj, event)
                elif event.type() == EventEnum.Leave:
                    self._handleLeaveEvent(obj, event)
                elif event.type() == EventEnum.MouseButtonPress:
                    return self._handleMousePressEvent(obj, event)
                elif event.type() == EventEnum.MouseButtonRelease:
                    return self._handleMouseReleaseEvent(obj, event)
                elif event.type() == EventEnum.User:
                    # handle enter & leave
                    if hasattr(event, '_pqi_is_enter'):
                        is_enter = event._pqi_is_enter
                        if is_enter:
                            self._handleEnterEvent(obj, event)
                        else:
                            self._handleLeaveEvent(obj, event)
                    # handle highlight
                    if hasattr(event, '_pqi_is_highlight'):
                        is_highlight = event._pqi_is_highlight
                        if is_highlight:
                            HighlightController.highlight(obj)
                        else:
                            HighlightController.unhighlight(obj)
                    # handle code exec
                    if hasattr(event, '_pqi_exec_code'):
                        code = event._pqi_exec_code
                        obj._pqi_exec(code)
            return False

    if IS_WINDOWS:
        class NativeEventListener(QtCore.QAbstractNativeEventFilter):
            """
            For some widgets that have overloaded the nativeEvent, mouse events may be intercepted earlier.
            Therefore, a NativeEventFilter needs to be implemented to prevent mouse events from being intercepted.
            """
            HTCLIENT = 1
            WM_NCHITTEST = 0x0084
            def nativeEventFilter(self, eventType, message):
                if not _is_inspect_enabled():
                    # If inspect is disabled, do not handle native events
                    return False, 0

                msg = wintypes.MSG.from_address(int(message))
                if msg.message == self.WM_NCHITTEST:
                    return True, self.HTCLIENT
                return False, 0
    elif IS_MACOS:
        class NativeEventListener(QtCore.QAbstractNativeEventFilter):
            """
            For macOS, when the window is not focused, the mouse event will not be triggered.
            Therefore, a NativeEventFilter needs to be implemented to obtain the mouse position
            and generate the corresponding enter and leave events.
            """
            __last_widget = None

            def nativeEventFilter(self, eventType, _):
                if not _is_inspect_enabled():
                    # If inspect is disabled, do not handle native events
                    return False, 0

                if eventType == 'mac_generic_NSEvent':
                    if QtGui.QGuiApplication.instance().focusWindow():
                        # If the window is focused, the event handle is not needed
                        self.__last_widget = None
                        return False, 0

                    locationInWindow = QtGui.QCursor.pos()
                    targetWidget = QtWidgets.QApplication.instance().widgetAt(locationInWindow)
                    if not targetWidget:
                        if self.__last_widget:
                            leaveEvent = QtCore.QEvent(EventEnum.User)
                            leaveEvent._pqi_is_enter = False
                            QtWidgets.QApplication.postEvent(self.__last_widget, leaveEvent)
                        self.__last_widget = None
                        return False, 0

                    if targetWidget != self.__last_widget:
                        # generate enter event
                        enterEvent = QtCore.QEvent(EventEnum.User)
                        enterEvent._pqi_is_enter = True
                        QtWidgets.QApplication.postEvent(targetWidget, enterEvent)
                        # generate leave event
                        if self.__last_widget:
                            leaveEvent = QtCore.QEvent(EventEnum.User)
                            leaveEvent._pqi_is_enter = False
                            QtWidgets.QApplication.postEvent(self.__last_widget, leaveEvent)
                        self.__last_widget = targetWidget
                return False, 0
    else:
        NativeEventListener = None

    def _initGlobalEventFilter():
        """ Initialize the global event filters when it does not exist """
        debugger = get_global_debugger()
        assert debugger is not None

        app = QtWidgets.QApplication.instance()
        if app is None:
            sys.stderr.write("QtWidgets.QApplication.instance() is None, stop attaching...")
            sys.stderr.flush()
            return

        # find one of the top-level widgets to move the event filter to the main thread
        topLevelWidgets = QtWidgets.QApplication.topLevelWidgets()
        if not topLevelWidgets:
            sys.stderr.write("Top-level widgets not found, stop attaching...")
            sys.stderr.flush()
            return

        topLevelWgt = topLevelWidgets[0]

        # Global event filter
        if debugger.global_event_filter is None:
            eventFilter = EventListener()
            # We need to move the event filter to the main thread
            eventFilter.moveToThread(topLevelWgt.thread())
            debugger.global_event_filter = eventFilter
            app.installEventFilter(eventFilter)

        # Global native event filter
        if debugger.global_native_event_filter is None and NativeEventListener is not None:
            nativeEventFilter = NativeEventListener()
            debugger.global_native_event_filter = nativeEventFilter
            app.installNativeEventFilter(nativeEventFilter)

    def _patchWidget(obj, *, attach=False):
        """ Install event listener and register widget to debugger """
        debugger = get_global_debugger()
        assert debugger is not None
        if not attach:
            # We use the Qt property system to mark the widget inspected
            #   because Python binding instance may change and lose the mark
            # To avoid the situation that attributes being dynamically added inside the `__init__` method
            #   and Qt inside directly executing the `event` method.
            # At this point, some custom classes may not have fully initialized
            #   and the event method can reference an uninitialized attribute.
            # So we use the QTimer and wait for `__init__` to finish.
            # ---
            # Bug Fixed 20240819: when the widget is deleted, the QTimer will not be executed.
            #   So we need to check if the widget is deleted before setting the property.
            # ---
            QtCore.QTimer.singleShot(0, lambda: obj.setProperty(_PQI_INSPECTED_PROP_NAME, True) if not isdeleted(obj) else None)
        else:
            # Attach thread may be different from the main thread,
            #   so the timer method will be invalid.
            # We just set the property directly because the widget has been initialized.
            obj.setProperty(_PQI_INSPECTED_PROP_NAME, True)
        # === register widget === #
        _register_widget(obj)

    def _new_QWidget_init(self, *args, **kwargs):
        self._original_QWidget_init(*args, **kwargs)
        if not ispycreated(self):
            # DO NOT install event listener for non-pycreated widget, because it may cause crash when exit
            return

        # === save stack when create === #
        frames = getStackFrame()
        setattr(self, '_pqi_stacks_when_create', frames)

        # Initialize the global filter when it does not exist
        _initGlobalEventFilter()

        # Patch widget
        _patchWidget(self)
        # Patch special widgets created by C++
        for specialMethod in ['viewport', 'tabBar', 'header']:
            if hasattr(self, specialMethod):
                p = getattr(self, specialMethod)
                if callable(p) and isinstance(p(), QtWidgets.QWidget):
                    for child in self.findChildren(QtWidgets.QWidget):
                        if not child.property(_PQI_INSPECTED_PROP_NAME):
                            _patchWidget(child)
                    break
        # for QAbstractSpinBox we should install event listener on its line edit
        if isinstance(self, QtWidgets.QAbstractSpinBox):
            _patchWidget(self.lineEdit())

    def _pqi_exec(self: QtWidgets.QWidget, code):
        debugger = get_global_debugger()
        try:
            f = StringIO()
            with redirect_stdout(f):
                exec(code, globals(), locals())
            if debugger is not None:
                debugger.notify_exec_code_result(f.getvalue())
        except Exception as e:
            if debugger is not None:
                debugger.notify_exec_code_error_message(str(e))

    def _notify_patch_success():
        _debugger = get_global_debugger()
        if _debugger is not None:
            _debugger.send_qt_patch_success_message()

    # ================================#
    #            ATTACH               #
    # ================================#
    def _patch_old_widgets_when_attached():
        # Initialize the global filter when beginning attach
        _initGlobalEventFilter()
        # Patch the existing widgets
        topLevelWidgets = QtWidgets.QApplication.topLevelWidgets()
        widgetsToPatch = collections.deque(topLevelWidgets)
        while widgetsToPatch:  # BFS traverse
            widget = widgetsToPatch.popleft()

            if isdeleted(widget) or not ispycreated(widget):
                continue

            # === patch widget ===
            _patchWidget(widget, attach=True)

            widgetsToPatch.extend(widget.findChildren(QtWidgets.QWidget))

    # 对于PyQt, 仅需patch基类QWidget即可
    # 但对于PySide, 则需要给所有的QWidget子类都打上补丁
    classesToPatch = QtWidgetClasses if qt_support_mode.startswith('pyside') else ['QWidget']

    for widgetClsName in classesToPatch:
        widgetCls = getattr(QtWidgets, widgetClsName)
        widgetCls._original_QWidget_init = widgetCls.__init__
        widgetCls.__init__ = _new_QWidget_init
        widgetCls._pqi_exec = _pqi_exec

    if is_attach:
        _patch_old_widgets_when_attached()

    pqi_log.info(f"pid {os.getpid()} patched.")
    _notify_patch_success()



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2024/8/20 15:25
# Description: Utils for logging
# ==============================================

class log_exception:
    def __init__(self, *, suppress=False):
        self.suppress = suppress

    def __enter__(self):
        from PyQtInspect._pqi_bundle import pqi_log
        self.pqi_log = pqi_log

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.pqi_log.error(f'Detect exception: {exc_val}', exc_info=True)
            if not self.suppress:
                return False
        return True



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/10/10 20:44
# Description: 
# ==============================================
import sys
import os

module_path = os.path.dirname(__file__)
sys.path.insert(0, os.path.dirname(module_path))

if __name__ == '__main__':
    import PyQtInspect.pqi

    from PyQtInspect.pqi import SetupHolder

    PyQtInspect.pqi.main()



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/10/9 19:20
# Description: 
# ==============================================
import typing

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtWidgets import QPushButton, QMenu, QWidgetAction, QApplication, QAction
from PyQt5.QtGui import QPixmap, QPainter, QFontMetrics, QColor, QBrush, QFont, QIcon
from PyQt5.QtCore import Qt, QRect, pyqtSignal, QPoint

from PyQtInspect.pqi_gui.children_menu_widget import ChildrenMenuWidget

ARROW_WIDTH = 17
SPACE_WIDTH = 8
ITEM_HEIGHT = 20


def clearLayout(layout):
    while layout.count():
        child = layout.takeAt(0)
        if child.widget():
            child.widget().deleteLater()


class HierarchyItem(QPushButton):
    arrowClicked = pyqtSignal()

    def __init__(self, parent, clsName: str, objName: str, widgetId: int, barWidget: 'HierarchyBar'):
        super().__init__(parent)
        self._barWidget = barWidget
        # === for paint === #
        self._pressed = False
        self._pressedText = False
        self._lastPressedText = False
        self._moveFlag = False
        self._menuShowed = False

        # === data === #
        self._clsName = clsName
        self._objName = objName
        self._widgetId = widgetId

        self._text = f"{clsName}{objName and f'#{objName}'}"

        self.m_font = parent.font()
        if self._text == "":
            self._textWidth = 0
            self.setFixedSize(ARROW_WIDTH, 21)
        else:
            fm = QFontMetrics(self.m_font)
            self._textWidth = fm.width(self._text) + SPACE_WIDTH
            self.setFixedSize(self._textWidth + ARROW_WIDTH, 21)

        self.m_normalIcon = QPixmap(":/icons/arrow_right.png")
        self.m_checkedIcon = QPixmap(":/icons/arrow_down.png")

        self.setMouseTracking(True)
        self.setCheckable(True)
        self.setToolTip(f"{self._text} (id {widgetId})")

        self.arrowClicked.connect(self._onArrowClicked)

    def getWidgetId(self) -> int:
        return self._widgetId

    def paintEvent(self, event):
        painter = QPainter(self)

        if self._moveFlag:
            painter.setPen(QColor(205, 232, 254))
            painter.fillRect(QRect(0, 0, self._textWidth + ARROW_WIDTH, ITEM_HEIGHT), QBrush(QColor(229, 243, 254)))
            painter.drawRect(QRect(0, 0, self._textWidth, ITEM_HEIGHT))
            painter.drawRect(QRect(self._textWidth, 0, ARROW_WIDTH, ITEM_HEIGHT))

        if self._pressed or self.isChecked():
            painter.setPen(QColor(153, 209, 255))
            painter.fillRect(QRect(0, 0, self._textWidth + ARROW_WIDTH, ITEM_HEIGHT), QBrush(QColor(204, 232, 255)))
            painter.drawRect(QRect(0, 0, self._textWidth, ITEM_HEIGHT))
            painter.drawRect(QRect(self._textWidth, 0, ARROW_WIDTH, ITEM_HEIGHT))

            painter.setPen(QColor(0, 0, 0))
            painter.setFont(self.m_font)
            painter.drawText(QRect(1, 1, self._textWidth, ITEM_HEIGHT), Qt.AlignCenter, self._text)

        else:
            painter.setPen(QColor(0, 0, 0))
            painter.setFont(self.m_font)
            painter.drawText(QRect(0, 0, self._textWidth, ITEM_HEIGHT), Qt.AlignCenter, self._text)

        # 绘制箭头
        if self._menuShowed:
            painter.drawPixmap(QRect(self._textWidth, 0, ARROW_WIDTH, ITEM_HEIGHT), self.m_checkedIcon,
                               self.m_checkedIcon.rect())
        else:
            painter.drawPixmap(QRect(self._textWidth, 0, ARROW_WIDTH, ITEM_HEIGHT), self.m_normalIcon,
                               self.m_normalIcon.rect())

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and self.hitButton(event.pos()):
            self._pressed = True
            if QRect(0, 0, self._textWidth, ITEM_HEIGHT).contains(event.pos()):
                self._pressedText = True
                self._lastPressedText = True
            elif QRect(self._textWidth, 0, ARROW_WIDTH, ITEM_HEIGHT).contains(event.pos()):
                self._lastPressedText = False
                self.update()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            if self._pressed:
                if self._pressedText:
                    self.setChecked(True)
                else:
                    self.arrowClicked.emit()

            self._pressedText = False
            self._pressed = False
            self.update()

    def enterEvent(self, event):
        self._moveFlag = True
        self.update()
        self._barWidget.notifyItemHovered(self)  # 向上层通知鼠标进入

    def leaveEvent(self, event):
        self._moveFlag = False
        self.update()

    def menuAboutToHide(self):
        self._menuShowed = False
        self._moveFlag = False
        self._pressed = False
        self.update()

    def _onArrowClicked(self):
        self.update()
        globalPoint = self.mapToGlobal(QPoint(0, 0))
        globalPoint.setX(globalPoint.x() + ARROW_WIDTH + self._textWidth - 30)
        globalPoint.setY(globalPoint.y() + ITEM_HEIGHT)
        self._menuShowed = True
        self._barWidget.notifyShowMenu(self, globalPoint.x(), globalPoint.y())


HIERARCHY_BAR_STYLE_SHEET = """
QScrollArea { background: transparent; border:none;}
"""


class HierarchyBarScrollArea(QtWidgets.QScrollArea):
    def __init__(self, parent):
        super().__init__(parent)
        self.setFixedHeight(21)
        self.setContentsMargins(0, 0, 0, 0)
        self.setObjectName("hierarchyBarScrollArea")
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    def wheelEvent(self, event):
        # 响应鼠标滚轮事件, 倒置事件的坐标系, 使得横向滚动条也能滚动
        newEvent = QtGui.QWheelEvent(event.pos(), event.globalPos(),
                                     QtCore.QPoint(event.pixelDelta().y(), event.pixelDelta().x()),
                                     QtCore.QPoint(event.angleDelta().y(), event.angleDelta().x()),
                                     event.buttons(), event.modifiers(), event.phase(), event.inverted())
        QtWidgets.QApplication.sendEvent(self.horizontalScrollBar(), newEvent)


class HierarchyBar(QtWidgets.QWidget):
    sigAncestorItemChanged = pyqtSignal(str)  # widgetId
    sigAncestorItemHovered = pyqtSignal(str)  # widgetId

    sigReqChildWidgetsInfo = pyqtSignal(str)  # widgetId
    sigChildMenuItemClicked = pyqtSignal(str)  # widgetId
    sigChildMenuItemHovered = pyqtSignal(str)  # widgetId

    sigMouseLeaveBarAndMenu = pyqtSignal()  # The mouse leaves both the bar and the popup menu.

    def __init__(self, parent):
        super().__init__(parent)
        self.setFixedHeight(21)

        self._buttonGroup = QtWidgets.QButtonGroup(self)
        self._buttonGroup.setExclusive(True)
        self._buttonGroup.buttonToggled.connect(self._onButtonToggled)  # 不能用clicked, 因为鼠标事件被拦截

        self._mainLayout = QtWidgets.QHBoxLayout(self)
        self._mainLayout.setContentsMargins(0, 0, 0, 0)
        self._mainLayout.setSpacing(2)

        self._goLeftButton = QtWidgets.QPushButton(self)
        self._goLeftButton.setFixedSize(16, 21)
        self._goLeftButton.setIcon(QIcon(":/icons/go_left.png"))
        self._goLeftButton.setObjectName("goLeftButton")
        self._goLeftButton.clicked.connect(self._onGoLeftButtonClicked)

        self._mainLayout.addWidget(self._goLeftButton)

        self._container = QtWidgets.QWidget(self)
        self._container.setObjectName("hierarchyBarContainer")

        self._containerLayout = QtWidgets.QHBoxLayout(self._container)
        self._containerLayout.setContentsMargins(0, 0, 0, 0)
        self._containerLayout.setSpacing(0)

        self._scrollArea = HierarchyBarScrollArea(self)
        self._scrollArea.setWidget(self._container)
        self._scrollArea.horizontalScrollBar().rangeChanged.connect(self._handleScrollButtonVisible)
        self._scrollArea.horizontalScrollBar().rangeChanged.connect(self._scrollToEndWhenRangeChanged)
        self._scrollArea.horizontalScrollBar().valueChanged.connect(self._handleScrollButtonEnabled)

        self._mainLayout.addWidget(self._scrollArea)

        self._goRightButton = QtWidgets.QPushButton(self)
        self._goRightButton.setFixedSize(16, 21)
        self._goRightButton.setIcon(QIcon(":/icons/go_right.png"))
        self._goRightButton.setObjectName("goRightButton")
        self._goRightButton.clicked.connect(self._onGoRightButtonClicked)

        self._mainLayout.addWidget(self._goRightButton)

        self._menu = QMenu(self)
        self._menu.setWindowFlags(Qt.Popup)
        self._menu.aboutToHide.connect(self._menuAboutToHide)
        action = QWidgetAction(self._menu)
        self._menuWidget = ChildrenMenuWidget(self, self._menu)
        action.setDefaultWidget(self._menuWidget)
        self._menu.addAction(action)

        self._menuWidget.sigClickChild.connect(self._handleChildMenuItemClicked)
        self._menuWidget.sigHoverChild.connect(self._handleChildMenuItemHovered)
        self._menuWidget.sigMouseLeave.connect(self._checkMouseLeave)

        self._curItemWithMenuShowed: typing.Optional[HierarchyItem] = None
        self._curCheckedItem: typing.Optional[HierarchyItem] = None
        self._widgetIdOfCurrMenu = -1
        self._lastHoveredChildItemWidgetId = -1

        self.setStyleSheet(HIERARCHY_BAR_STYLE_SHEET)
        self._handleScrollButtonVisible()

    def _onGoLeftButtonClicked(self):
        self._scrollArea.horizontalScrollBar().setValue(
            self._scrollArea.horizontalScrollBar().value() - self._scrollArea.width() // 2
        )

    def _onGoRightButtonClicked(self):
        self._scrollArea.horizontalScrollBar().setValue(
            self._scrollArea.horizontalScrollBar().value() + self._scrollArea.width() // 2
        )

    def _handleScrollButtonVisible(self, *_):
        if self._container.width() > self._scrollArea.width():
            self._goLeftButton.show()
            self._goRightButton.show()
        else:
            self._goLeftButton.hide()
            self._goRightButton.hide()

    def _handleScrollButtonEnabled(self, value):
        if value == 0:
            self._goLeftButton.setEnabled(False)
        else:
            self._goLeftButton.setEnabled(True)

        if value == self._scrollArea.horizontalScrollBar().maximum():
            self._goRightButton.setEnabled(False)
        else:
            self._goRightButton.setEnabled(True)

    def setData(self,
                ancestorClsNameList: typing.List[str],
                ancestorObjNameList: typing.List[str],
                ancestorWidgetIdList: typing.List[int]):
        # clear old data
        clearLayout(self._containerLayout)
        self._buttonGroup.buttons().clear()
        self._curItemWithMenuShowed = None
        self._curCheckedItem = None

        for clsName, objName, widgetId in zip(ancestorClsNameList, ancestorObjNameList, ancestorWidgetIdList):
            item = HierarchyItem(self, clsName, objName, widgetId, self)
            self._buttonGroup.addButton(item)
            self._containerLayout.addWidget(item)
        self._containerLayout.addStretch()

        if self._buttonGroup.buttons():  # not empty, set last item checked by default
            lastItem = self._buttonGroup.buttons()[-1]
            # 之所以在这里设置, 是因为避免数据变更后emit sigAncestorItemChanged信号
            # 导致重复获取数据
            self._curCheckedItem = lastItem
            lastItem.setChecked(True)

        self.update()

    def _scrollToEndWhenRangeChanged(self, _min, _max):
        self._scrollArea.horizontalScrollBar().setValue(_max)

    def notifyShowMenu(self, itemWidget: HierarchyItem, posX: int, posY: int):
        self._curItemWithMenuShowed = itemWidget

        self._menu.move(posX, posY)
        if self._widgetIdOfCurrMenu != itemWidget.getWidgetId():
            # 脏数据
            self._menuWidget.setLoading()
            self.sigReqChildWidgetsInfo.emit(str(itemWidget.getWidgetId()))
        self._menu.show()

    def notifyHideMenu(self, itemWidget: HierarchyItem):
        if self._curItemWithMenuShowed == itemWidget:
            self._menu.hide()
            self._curItemWithMenuShowed = None

    def _menuAboutToHide(self):
        if self._curItemWithMenuShowed:
            self._curItemWithMenuShowed.menuAboutToHide()

    def notifyItemHovered(self, itemWidget: HierarchyItem):
        self.sigAncestorItemHovered.emit(str(itemWidget.getWidgetId()))

    def _onButtonToggled(self, btn, checked):
        if checked and btn != self._curCheckedItem:
            self._curCheckedItem = btn
            self.sigAncestorItemChanged.emit(str(btn.getWidgetId()))

    def setMenuData(self,
                    widgetId: int,
                    childClsNameList: typing.List[str],
                    childObjNameList: typing.List[str],
                    childWidgetIdList: typing.List[int]):
        if self._curItemWithMenuShowed is None or widgetId != self._curItemWithMenuShowed.getWidgetId():
            # 脏数据
            return

        self._widgetIdOfCurrMenu = widgetId
        self._menuWidget.setMenuData(childClsNameList, childObjNameList, childWidgetIdList)

    def _handleChildMenuItemHovered(self, widgetId: str):
        if widgetId == "-1":
            return

        if self._lastHoveredChildItemWidgetId == widgetId:
            return

        self._lastHoveredChildItemWidgetId = widgetId
        self.sigChildMenuItemHovered.emit(widgetId)

    def _handleChildMenuItemClicked(self, widgetId: str):
        self._menu.hide()
        if widgetId != "-1":
            self.sigChildMenuItemClicked.emit(widgetId)

    def leaveEvent(self, event):
        super().enterEvent(event)
        self._checkMouseLeave()

    def _checkMouseLeave(self):
        """ Check if the mouse has left the bar and menu.
        If so, emit a signal.
        """
        if not self.underMouse() and not self._menuWidget.underMouse():
            self.sigMouseLeaveBarAndMenu.emit()

    def clearData(self):
        clearLayout(self._containerLayout)
        self._buttonGroup.buttons().clear()
        self._curItemWithMenuShowed = None
        self._curCheckedItem = None
        self._widgetIdOfCurrMenu = -1
        self._lastHoveredChildItemWidgetId = -1


class TestWindow(QtWidgets.QMainWindow):
    def __init__(self, parent):
        super().__init__(parent)
        self.setMinimumSize(500, 500)
        self._widget = HierarchyBar(self)
        # self._giftInfoWidget.setOrder(1)
        self.setCentralWidget(self._widget)
        self._widget.setData(
            ["QWidget", "QWidget", "QWidget", "QWidget", "QWidget", "QWidget", "QWidget", "QWidget", "QWidget"],
            ["testWidget", "testWidget", "testWidget", "testWidget", "testWidget", "testWidget", "testWidget",
             "testWidget", "testWidget"],
            [1, 2, 3, 4, 5, 6, 7, 8, 9]
        )
        self._widget.sigAncestorItemHovered.connect(lambda widgetId: print(f"hovered {widgetId}"))
        self._widget.sigAncestorItemChanged.connect(lambda widgetId: print(f"changed {widgetId}"))
        self._widget.sigChildMenuItemClicked.connect(lambda widgetId: print(f"triggered {widgetId}"))
        self._widget.sigChildMenuItemHovered.connect(lambda widgetId: print(f"hovered {widgetId}"))

        self._setMenuDataBtn = QtWidgets.QPushButton(self)
        self._setMenuDataBtn.setText("setMenuData")
        self._setMenuDataBtn.move(100, 100)
        self._setMenuDataBtn.clicked.connect(self._onSetMenuDataBtnClicked)

    def _onSetMenuDataBtnClicked(self):
        self._widget.setMenuData(9,
                                 ["QWidget"] * 30,
                                 ["testWidget"] * 30,
                                 [i for i in range(30)])


if __name__ == '__main__':
    import sys

    app = QtWidgets.QApplication(sys.argv)
    import PyQtInspect.pqi_gui._pqi_res.resources

    parentWin = TestWindow(None)
    parentWin.show()

    sys.exit(app.exec_())



# -*- encoding:utf-8 -*-

import contextlib
import io
import typing

from PyQt5 import QtWidgets, QtGui, QtCore

from PyQtInspect.pqi_gui._pqi_res import get_icon
from PyQtInspect.pqi_gui.components.simple_kv_line_edit import SimpleSettingLineEdit
import PyQtInspect.pqi_gui.data_center as DataCenter


PYTHON_NAMES = ['python', ]


def _getAllPySubprocess(parent_pid):
    import psutil

    parent_process = psutil.Process(parent_pid)
    children = parent_process.children(recursive=True)
    for child in children:
        for pyName in PYTHON_NAMES:
            if pyName in child.name():
                yield child


class PidLineEdit(SimpleSettingLineEdit):
    sigAttachButtonClicked = QtCore.pyqtSignal()
    sigAttachAllPySubprocessButtonClicked = QtCore.pyqtSignal()  # parent_pid

    def __init__(self, parent):
        super().__init__(parent, "Pid: ")

        self._attachButton = QtWidgets.QPushButton(self)
        self._attachButton.setText("Attach")
        self._attachButton.setFixedHeight(30)
        self._attachButton.clicked.connect(self.sigAttachButtonClicked)

        self._layout.addWidget(self._attachButton)

        self._attachAllPySubButton = QtWidgets.QPushButton(self)
        self._attachAllPySubButton.setText("Attach All Python Subprocesses")
        self._attachAllPySubButton.setFixedHeight(30)
        self._attachAllPySubButton.clicked.connect(self.sigAttachAllPySubprocessButtonClicked)

        self._layout.addWidget(self._attachAllPySubButton)
        self._initGrabAction()

    def _initGrabAction(self):
        """ Initialize the grab action for getting pid from cursor.

        @note: Only available on Windows.
        """
        import sys
        if sys.platform == "win32":
            def _tryGrab():
                import wingrab
                pid = wingrab.grab()
                self._valueLineEdit.setText(str(pid))

            self._grabAction = self._valueLineEdit.addAction(QtGui.QIcon(":/cursors/cursor.png"),
                                                             QtWidgets.QLineEdit.TrailingPosition)
            self._grabAction.triggered.connect(_tryGrab)
            self._grabAction.setToolTip("Get pid from cursor")


class AttachInfoTextBrowser(QtWidgets.QTextBrowser):
    def write(self, text):
        self.append(text)


class AttachWorker(QtCore.QObject):
    sigStdOut = QtCore.pyqtSignal(str)
    sigAttachFinished = QtCore.pyqtSignal()
    sigAttachError = QtCore.pyqtSignal(str)

    class _StdOutGetter(io.StringIO):
        def __init__(self, worker):
            super().__init__()
            self._worker = worker

        def write(self, text):
            super().write(text)
            self._worker.sigStdOut.emit(f"({self._worker.pidToAttach}) {text}")

    def __init__(self, parent, pidToAttach: int):
        super().__init__(parent)
        self._pidToAttach = pidToAttach

    @property
    def pidToAttach(self):
        return self._pidToAttach

    def doWork(self):
        from PyQtInspect.pqi_attach.attach_pydevd import main as attach_main_func

        with contextlib.redirect_stdout(AttachWorker._StdOutGetter(self)):
            try:
                print('Attaching...')
                attach_main_func(
                    {
                        'port': DataCenter.instance.port,
                        'pid': self._pidToAttach,
                        'host': '127.0.0.1',
                        'protocol': '', 'debug_mode': ''
                    }
                )
                # print('==================')

                self.sigAttachFinished.emit()
            except Exception as e:
                print(f'Attach Error: {e}\n')
                self.sigAttachError.emit(str(e))


class AttachMultipleWorker(QtCore.QObject):
    sigStdOut = QtCore.pyqtSignal(str)
    sigAttachFinished = QtCore.pyqtSignal(int)  # errCount

    class _StdOutGetter(io.StringIO):
        def __init__(self, worker):
            super().__init__()
            self._worker = worker

        def write(self, text):
            super().write(text)
            self._worker.sigStdOut.emit(text)

    def __init__(self, parent, pidsToAttach: typing.List[int]):
        super().__init__(parent)
        self._pidsToAttach = pidsToAttach

        self._subWorkers = []
        self._subThreads = set()
        self._errCount = 0

    def _onSubThreadQuit(self, thread):
        thread.quit()
        self._subThreads.remove(thread)
        if not self._subThreads:
            self.sigAttachFinished.emit(self._errCount)

    def _onSubWorkerAttachError(self, pid, errMsg):
        with contextlib.redirect_stdout(AttachMultipleWorker._StdOutGetter(self)):
            print(f'Attach {pid} Error: {errMsg}\n')
            self._errCount += 1

    def doWork(self):
        for pid in self._pidsToAttach:
            worker = AttachWorker(None, pid)
            self._subWorkers.append(worker)
            thread = QtCore.QThread(None)
            worker.moveToThread(thread)

            thread.started.connect(worker.doWork)
            thread.finished.connect(lambda _thread=thread: self._onSubThreadQuit(_thread))

            worker.sigAttachFinished.connect(thread.quit)
            worker.sigAttachError.connect(lambda msg, _pid=pid: self._onSubWorkerAttachError(_pid, msg))
            worker.sigStdOut.connect(self.sigStdOut)

            self._subThreads.add(thread)
            thread.start()


class AttachWindow(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowFlags(QtCore.Qt.Window | QtCore.Qt.CustomizeWindowHint | QtCore.Qt.WindowCloseButtonHint)
        self.setWindowTitle("Attach to Process")
        self.setWindowIcon(get_icon())
        self.resize(500, 300)

        self._mainLayout = QtWidgets.QVBoxLayout(self)
        self._mainLayout.setContentsMargins(10, 10, 10, 10)
        self._mainLayout.setSpacing(5)
        self._mainLayout.addSpacing(4)

        self._pidLine = PidLineEdit(self)
        self._pidLine.sigAttachButtonClicked.connect(self._onAttachButtonClicked)
        self._pidLine.sigAttachAllPySubprocessButtonClicked.connect(self._onAttachAllPySubprocessButtonClicked)
        self._mainLayout.addWidget(self._pidLine)

        self._mainLayout.addSpacing(4)

        self._consoleOutputTextBrowser = AttachInfoTextBrowser(self)
        self._consoleOutputTextBrowser.setReadOnly(True)
        self._mainLayout.addWidget(self._consoleOutputTextBrowser)

        self._thread = None
        self._worker = None

    def _onAttachButtonClicked(self):
        try:
            self._tryAttachToProcess(int(self._pidLine.getValue()))
        except ValueError:
            # todo 如果pid不存在呢??
            QtWidgets.QMessageBox.critical(self, "Error", "Invalid Pid!")
            return

    def _onAttachAllPySubprocessButtonClicked(self):
        try:
            self._tryAttachToAllPySubprocess(int(self._pidLine.getValue()))
        except ValueError:
            # todo 如果pid不存在呢??
            QtWidgets.QMessageBox.critical(self, "Error", "Invalid Pid!")
            return

    def _doWorkerImpl(self, worker):
        worker.sigAttachFinished.connect(self._thread.quit)
        worker.sigAttachError.connect(self._onAttachError)
        worker.sigStdOut.connect(self._consoleOutputTextBrowser.write)
        self._thread.started.connect(worker.doWork)
        self._thread.finished.connect(lambda: self._pidLine.setEnabled(True))

    def _tryAttachToProcess(self, pid: int):
        """ Attach放在这个控件实现吧 """
        if self._thread is not None and self._thread.isRunning():  # guard
            QtWidgets.QMessageBox.information(self, "Info", "Already Attaching")
            return

        from PyQtInspect.pqi_attach.attach_pydevd import main as attach_main_func

        self._pidLine.setEnabled(False)

        self._worker = AttachWorker(None, pid)

        self._thread = QtCore.QThread(None)
        # must move to thread before connect
        self._worker.moveToThread(self._thread)

        self._worker.sigStdOut.connect(self._consoleOutputTextBrowser.write)
        self._worker.sigAttachError.connect(self._onAttachError)
        self._thread.started.connect(self._worker.doWork)
        self._thread.finished.connect(lambda: self._pidLine.setEnabled(True))

        self._worker.sigAttachFinished.connect(self._thread.quit)
        self._thread.start()

    def _tryAttachToAllPySubprocess(self, parent_pid: int):
        if self._thread is not None and self._thread.isRunning():  # guard
            QtWidgets.QMessageBox.information(self, "Info", "Already Attaching")
            return

        childPids = [childProcess.pid for childProcess in _getAllPySubprocess(parent_pid)]
        if not childPids:
            QtWidgets.QMessageBox.information(self, "Info", "No Python Subprocess Found")
            return

        self._pidLine.setEnabled(False)

        self._thread = QtCore.QThread(None)
        self._worker = AttachMultipleWorker(None, childPids)
        self._worker.moveToThread(self._thread)
        self._worker.sigStdOut.connect(self._consoleOutputTextBrowser.write)
        self._thread.started.connect(self._worker.doWork)
        self._thread.finished.connect(lambda: self._pidLine.setEnabled(True))

        self._worker.sigAttachFinished.connect(self._thread.quit)
        self._thread.start()

    def _onAttachError(self, errMsg):
        QtWidgets.QMessageBox.critical(self, "Error", errMsg)
        self._pidLine.setEnabled(True)


if __name__ == '__main__':
    import sys

    app = QtWidgets.QApplication(sys.argv)
    window = AttachWindow()
    window.show()
    sys.exit(app.exec())



import os

from PyQt5 import QtWidgets, QtCore, QtGui
from PyQtInspect.pqi_gui.settings import getPyCharmPath, findDefaultPycharmPath, setPyCharmPath


class CreateStacksListWidget(QtWidgets.QListWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.setObjectName("stacksListWidget")
        self.setMinimumHeight(200)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)

    def setStacks(self, stacks: list):
        self.clear()
        for index, stack in enumerate(stacks):
            fileName = stack.get("filename", "")
            normalizedFileName = os.path.normpath(fileName)
            isSrc = os.path.exists(normalizedFileName)

            lineNo = stack.get("lineno", "")
            funcName = stack.get("function", "")
            item = QtWidgets.QListWidgetItem()
            item.setText(f"{index + 1}. {'' if isSrc else '<?> '}File {normalizedFileName}, line {lineNo}: {funcName}")
            # set property
            item.setData(QtCore.Qt.UserRole, (isSrc, normalizedFileName, lineNo))
            self.addItem(item)

    def clearStacks(self):
        self.clear()

    # double click to open file
    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mousePressEvent(event)
        if event.button() == QtCore.Qt.LeftButton:
            item = self.itemAt(event.pos())
            if item is not None:
                isSrc, fileName, lineNo = item.data(QtCore.Qt.UserRole)
                if isSrc:
                    if fileName:
                        self.openFile(fileName, lineNo)
                else:  # we need to map to our local directory
                    # todo add a dialog to ask user to map the file
                    ...

    def findPycharm(self):
        pycharmPath = getPyCharmPath()
        if not pycharmPath:
            pycharmPath = findDefaultPycharmPath()
            if pycharmPath:
                setPyCharmPath(pycharmPath)
        return pycharmPath

    def openFile(self, fileName: str, lineNo: int):
        # open in Pycharm
        import subprocess
        pycharm = self.findPycharm()
        if pycharm:
            try:
                subprocess.Popen(f'"{pycharm}" --line {lineNo} "{fileName}"', shell=True)
            except Exception as e:
                # message box
                QtWidgets.QMessageBox.critical(self, "Error", f"Error occurred when opening file: {e}")
        else:
            QtWidgets.QMessageBox.critical(self, "Error", "Pycharm not found")
        # subprocess.Popen(f"pycharm64.exe --line {lineNo} {fileName}")



# -*- coding:utf-8 _*-

import logging
import os
import pathlib
import time

__logger_cache = {}
_PROGRAM_DIR_PATH = pathlib.Path.home() / '.PyQtInspect'
_LOG_DIR_PATH = _PROGRAM_DIR_PATH / 'log'
if not _LOG_DIR_PATH.exists():
    _LOG_DIR_PATH.mkdir(parents=True)


def getLogger(logger_name='PyQtInspect', console_log_level=logging.INFO, file_log_level=logging.INFO):
    if logger_name in __logger_cache:
        return __logger_cache[logger_name]

    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    # 1. Console
    sh = logging.StreamHandler()
    sh.setLevel(console_log_level)
    # 2. File
    rq = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime((time.time())))
    log_name = f'{logger_name}_{os.getpid()}_{rq}.log'
    log_path = _LOG_DIR_PATH.joinpath(log_name)
    fh = logging.FileHandler(log_path)
    fh.setLevel(file_log_level)

    # Format
    shFormatter = logging.Formatter(
        '[PyQtInspect][%(asctime)s][%(filename)s: line %(lineno)d][%(levelname)s] %(message)s'
    )
    sh.setFormatter(shFormatter)

    fhFormatter = logging.Formatter('[%(asctime)s][%(filename)s: line %(lineno)d][%(levelname)s] %(message)s')
    fh.setFormatter(fhFormatter)

    # Add handlers
    logger.addHandler(fh)
    logger.addHandler(sh)

    # Monkey patch for dynamic log level setting
    def set_console_log_level(level):
        if sh.level == level:
            return
        sh.setLevel(level)

    def set_file_log_level(level):
        if fh.level == level:
            return
        fh.setLevel(level)

    logger.set_console_log_level = set_console_log_level
    logger.set_file_log_level = set_file_log_level

    __logger_cache[logger_name] = logger
    return logger
